/*
  interact: model fault interactions using dislocations in a 
            halfspace
  (C) Thorsten Becker, becker@eps.harvard.edu

  $Id: divide_fault_in_patches.c,v 1.18 2004/03/16 01:11:05 becker Exp becker $
*/
#include "properties.h"
#include "interact.h"



void create_patches(int flt, struct flt *fault,struct flt **patch,
		    int *nrpatches,int *seg,my_boolean circular,
		    my_boolean calculate_base_vecs,COMP_PRECISION *td,
		    COMP_PRECISION srand, COMP_PRECISION drand,
		    long *seed)	/* gaussian randomization */
{
  int segn[2];
  determine_segments(seg,segn,(fault+flt),circular,td);
  divide_fault_in_patches(flt,fault,patch,nrpatches,segn,circular,
			  calculate_base_vecs,srand,drand,seed);
}
/*

  determine the number of subdivisions of a fault patch
  
*/

void determine_segments(int *seg,int *segn, struct flt *fault,
			my_boolean circular,COMP_PRECISION *td)
{
  int j;
  for(j=0;j<2;j++){
    if(td[j]>0){
      segn[j]=(int)((j==0?fault->l:fault->w)/td[j]+0.5);
      if(segn[j]<1)
	segn[j]=1;
      if(circular)
	if((segn[j] % 2) == 0)
	  segn[j]++;
    }else{
      segn[j]=seg[j];
    }
  } 
}

/* 


subdivide a rectangular surface into sub-fault patches


 */

void divide_fault_in_patches(int flt,struct flt *fault,
			     struct flt **patch,
			     int *nrpatches,int *seg,
			     my_boolean circular,
			     my_boolean calculate_base_vecs,
			     COMP_PRECISION srand, COMP_PRECISION drand,
			     long *seed)
{
  int i,j,mi,mj,irad,old_nrpatches,
    added_patches,pcnt;
  COMP_PRECISION dx[3],dy[3],alpha,sin_dip,cos_dip,
    corner[4][3],l2,w2,newl,neww,tmpd1,tmpd2,newarea,
    l,w;
  my_boolean randomize;
  randomize = ((srand>0)||(drand>0))?(TRUE):(FALSE);
  
  added_patches=seg[0]*seg[1];
  if(!added_patches){
    fprintf(stderr,"divide_fault_in_patches: nm %i/%i error\n",
	    seg[0],seg[1]);
    exit(-1);
  }
  // increase patches structure
  old_nrpatches= *nrpatches;
  if(! *nrpatches){
    *nrpatches += added_patches;
    *patch = (struct flt *)calloc(*nrpatches,sizeof(struct flt));
    if(! *patch){
      fprintf(stderr,"divide_fault_in_patches: mem error, nr patches: %i\n",*nrpatches);
      exit(-1);
    }
  }else{
    *nrpatches += added_patches;
    *patch = (struct flt *)realloc(*patch,*nrpatches*sizeof(struct flt));
    if(! *patch){
      fprintf(stderr,"divide_fault_in_patches: mem error, nr patches: %i\n",*nrpatches);
      exit(-1);
    }
  }
  // if base vecs not yet initialized, do now
  if(calculate_base_vecs){
    
    alpha = 90.0 - (COMP_PRECISION)fault[flt].strike;
    my_sincos_deg(&fault[flt].sin_alpha,&fault[flt].cos_alpha,alpha);
    my_sincos_deg(&sin_dip,&cos_dip,(COMP_PRECISION)fault[flt].dip);
    calc_base_vecs(fault[flt].t_strike,fault[flt].normal,fault[flt].t_dip,
		   fault[flt].sin_alpha,fault[flt].cos_alpha,sin_dip,cos_dip);
  }
  // obtain corner points of original patch
  calculate_corners(corner,(fault+flt),&l,&w);
  // new patch geometry
  newl = l / (COMP_PRECISION)seg[0];
  neww = w / (COMP_PRECISION)seg[1];
  l2 = newl * 2.0;
  w2 = neww * 2.0;
  newarea = l2 * w2;
  for(i=0;i<3;i++){
    dx[i] = fault[flt].t_strike[i]*l2;
    dy[i] = fault[flt].t_dip[i]*w2;
  }
  // assign common specs by copying them over from the 
  // fault to the patches
  for(i=old_nrpatches;i< *nrpatches;i++){
    fltcp((fault+flt),(*patch+i));

    if(randomize){		/* randomize this patch? */
      randomize_strike_dip(srand,drand,(*patch+i),seed);
      /* need to recompute basis vectors */
      alpha = 90.0 - (COMP_PRECISION)(*patch+i)->strike;
      my_sincos_deg(&((*patch+i)->sin_alpha),&((*patch+i)->cos_alpha),alpha);
      my_sincos_deg(&sin_dip,&cos_dip,(COMP_PRECISION)(*patch+i)->dip);
      calc_base_vecs((*patch+i)->t_strike,(*patch+i)->normal,(*patch+i)->t_dip,
		     (*patch+i)->sin_alpha,(*patch+i)->cos_alpha,sin_dip,cos_dip);
    }
    (*patch+i)->area = newarea;
    (*patch+i)->w = neww;
    (*patch+i)->l = newl;
  }
  if(!circular){/* divide fault patch in rectangular patches */
    /* 
       
    go left right, then top down

    */

    pcnt = old_nrpatches;
    for(j=seg[1]-1;j >= 0;j--){
      for(i=0;i<seg[0];i++){
	/* calculate position of patch in fault */
	get_flt_location((*patch + pcnt),dx,dy,&corner[0][0],i,j);
	pcnt++;
      }
    }
  }else{
    /* 
       divide fault patch in circular patches 
       lk and wk are the same since n and m are equal
    */
    if((seg[0] % 2 == 0)||(seg[1] % 2 == 0)){
      fprintf(stderr,"divide_fault_in_patches: need odd numbers for circular patches\n");
      exit(-1);
    }
    mi=(seg[1]+1)/2;mj=(seg[0]+1)/2;
    irad=MIN(mi,mj);
    for(pcnt=old_nrpatches,j=0;j<seg[0];j++){
      for(i=0;i<seg[1];i++){
	tmpd1=(COMP_PRECISION)(i-mi); 
	tmpd2=(COMP_PRECISION)(j-mj);
	if(hypot(tmpd1,tmpd2) > irad)
	  continue;
	/* calculate position of patch in fault */
	get_flt_location((*patch + pcnt),dx,dy,&corner[0][0],i,j);
	pcnt++;
      }
    }
  }
  // resize patch array if we did not need all the new patches
  if(pcnt<added_patches){
    *nrpatches = old_nrpatches + pcnt;
    *patch = (struct flt *)realloc(*patch,*nrpatches * sizeof(struct flt));
  }
}

/* 

get the center point location of a sub fault patch for strike spacing
dx, dip spacing dy, and strike index i, as well as dip index j 

*/
void get_flt_location(struct flt *fault,COMP_PRECISION *dx,COMP_PRECISION *dy,
		      COMP_PRECISION *corner,int i, int j)
{
  int k;
  for(k=0;k<3;k++){
    fault->x[k]  = corner[k]+(0.5+(COMP_PRECISION)i)*dx[k];
    fault->x[k] += (0.5+(COMP_PRECISION)j)*dy[k];
    if(fabs(fault->x[k])<EPS_COMP_PREC)
      fault->x[k]=0.0;
  }
}  

void randomize_strike_dip(COMP_PRECISION srand,
			  COMP_PRECISION drand,struct flt *fault,
			  long *seed)
{
  double strike,dip;
  /* compute Gaussian deviation with srand drand std */
  strike = fault->strike + mygauss_randnr(srand,seed); 
  dip    = fault->dip    + mygauss_randnr(drand,seed); 
  check_angles(&dip,&strike);
  fault->strike = strike;
  fault->dip = dip;
}
   
		      

/*

################################################################################
#                                                                              #
#  interact: model fault interactions using dislocations in an elastic         #
#            medium                                                            #
#	                                                                       #
#    Copyright (C) Thorsten W. Becker 2000 - 2002                              #
#                                                                              #
#    Unteract uses dc3d.f as provided by Y. Okada as in Okada (BSSA, 1992) and #
#    linear algebra routines from SLATEC, LAPACK, and EISPACK.                 #
#    Might contain material copyrighted by others (e.g. Numerical Recipes etc) #
#                                                                              #
#  If you use interact in real work, a citation to use would be:               #
#                                                                              #
#  @Article{becker02c,                                                         #
#    author = 	 {Becker, T. W. and Schott, B.},                               #
#    title = 	 {On boundary-element models of elastic fault interaction      #
#                  (abstract)},                                                # 
#    journal = 	 {EOS Trans. AGU},                                             #
#    year = 	 2002,                                                         #
#    volume =	 83,                                                           #
#    number =	 47,                                                           #
#    pages =	 {NG62A-0925}                                                  #
#  }                                                                           #
#                                                                              #
#                                                                              #
#    This program is free software; you can redistribute it and/or modify      #
#    it under the terms of the GNU General Public License as published by      #
#    the Free Software Foundation; either version 2 of the License, or         #
#    (at your option) any later version.                                       #
#                                                                              #
#    This program is distributed in the hope that it will be useful,           #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of            #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
#    GNU General Public License for more details.                              #
#                                                                              #
#    In addition, we strongly suggest that iGMT users comply with the goals    #
#    as expressed in the Student Pugwash Pledge (www.spusa.org/pugwash/).      #
#                                                                              #
#    You should have received a copy of the GNU General Public License         #
#    along with this program; see the file COPYING.  If not, write to          #
#    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,          #
#    Boston, MA 02111-1307, USA.                                               #
#                                                                              #
################################################################################

*/

