#include "interact.h"
#include "properties.h"
/*
 
reads in sets of four points in 3-D and converts to patch format

uses the fit_plane and points2patch subroutines, both are in
fit_plane.c

input format:

p_1^x p_1^y p_1^z
p_2^x p_2^y p_2^z
p_3^x p_3^y p_3^z
p_4^x p_4^y p_4^z

...


repeated for every rectangle. points have to be in finite element
ordering (counterclockwise)

the opmode flag determines the output format, see interact.h

$Id: points2patch.c,v 1.8 2003/07/02 07:57:34 becker Exp $

*/


int main(int argc, char **argv)
{
  
  struct flt fault[1],*patch;
  struct med *medium;
  COMP_PRECISION x[12],dx;
  int i,opmode=PATCH_OUT_MODE,nrpatches,seg[2];
  long seed = -1;
  my_boolean adjust_area=TRUE;
  medium=(struct med *)calloc(1,sizeof(struct med));

  dx = 1.5;

  if(argc > 3){
    fprintf(stderr,"%s: usage:\n\t%s [adjust_area, %i] [dx, %g]\n\tread in four 3-D points per line from stdin\n",
	    argv[0],argv[0],adjust_area,dx);
    fprintf(stderr,"\tif adjust_area is set, will attempt to make input and output same area\n");
    fprintf(stderr,"\tformat:\n\tx1_x x1_y x1_z x2_x x2_y x2_z ...\n\n");
    fprintf(stderr,"\tto form a regular quad, points have to be in FE (CCW) ordering\n");
    fprintf(stderr,"\tassumes that patch can be described by dip and strike only, no rake!\n");
    fprintf(stderr,"\twrites patch format to stdout\n");
    fprintf(stderr,"\tdx (%g) will subdivide the faults into patches with dx width/length\n",dx);
    exit(-1);
  }
  if(argc >= 2){
    sscanf(argv[1],"%i",&i);
    adjust_area=(my_boolean)i;
  }
  if(argc > 2)
    sscanf(argv[2],"%lf",&dx);
  
  fprintf(stderr,"%s: reading points from %s, writing patch to stdout,",
	  argv[0],"stdin");
  if(adjust_area)
    fprintf(stderr," adjusting area");
  else
    fprintf(stderr," area unadjusted");
  fprintf(stderr,", spacing %g\n",dx);

  medium->nrflt=0;
  nrpatches=0;
  while(fscanf(stdin,TWELVE_CP_FORMAT,
	       &x[  X],&x[  Y],&x[  Z],
	       &x[3+X],&x[3+Y],&x[3+Z],
	       &x[6+X],&x[6+Y],&x[6+Z],
	       &x[9+X],&x[9+Y],&x[9+Z]) == 12){
    for(i=0;i<4;i++)
      if(x[i*3+Z] > 0.0){
	fprintf(stderr,"%s: rectangle %i: point %i: z should be <= 0 (%g, %g, %g)\n",
		argv[0],medium->nrflt+1,i+1,x[i*3+X],x[i*3+Y],x[i*3+Z]);
	exit(-1);
      }
    points2patch(fault,x,adjust_area);
    fault[0].group = medium->nrflt;
    /* hack for now */
    seg[0] = 1 + fault[0].l/dx;
    seg[1] = 1 + fault[0].w/dx;
    /* subdivide */
    divide_fault_in_patches(0,fault,&patch,&nrpatches,
			    seg,FALSE,TRUE,0,0,&seed);
    medium->nrflt ++;
    
  }
  for(i=0;i<nrpatches;i++){
    print_patch_geometry_and_bc(i,patch,opmode,0.0,
				FALSE,stdout,FALSE);
  }
  fprintf(stderr,"%s: read %i sets of four points\n",
	  argv[0],medium->nrflt);
  fprintf(stderr,"%s: produced total %i faults with %i patches\n",
	  argv[0],medium->nrflt,nrpatches);
  exit(0);
}






/*

################################################################################
#                                                                              #
#  interact: model fault interactions using dislocations in an elastic         #
#            medium                                                            #
#	                                                                       #
#    Copyright (C) Thorsten W. Becker 2000 - 2002                              #
#                                                                              #
#    Unteract uses dc3d.f as provided by Y. Okada as in Okada (BSSA, 1992) and #
#    linear algebra routines from SLATEC, LAPACK, and EISPACK.                 #
#    Might contain material copyrighted by others (e.g. Numerical Recipes etc) #
#                                                                              #
#  If you use interact in real work, a citation to use would be:               #
#                                                                              #
#  @Article{becker02c,                                                         #
#    author = 	 {Becker, T. W. and Schott, B.},                               #
#    title = 	 {On boundary-element models of elastic fault interaction      #
#                  (abstract)},                                                # 
#    journal = 	 {EOS Trans. AGU},                                             #
#    year = 	 2002,                                                         #
#    volume =	 83,                                                           #
#    number =	 47,                                                           #
#    pages =	 {NG62A-0925}                                                  #
#  }                                                                           #
#                                                                              #
#                                                                              #
#    This program is free software; you can redistribute it and/or modify      #
#    it under the terms of the GNU General Public License as published by      #
#    the Free Software Foundation; either version 2 of the License, or         #
#    (at your option) any later version.                                       #
#                                                                              #
#    This program is distributed in the hope that it will be useful,           #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of            #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
#    GNU General Public License for more details.                              #
#                                                                              #
#    In addition, we strongly suggest that iGMT users comply with the goals    #
#    as expressed in the Student Pugwash Pledge (www.spusa.org/pugwash/).      #
#                                                                              #
#    You should have received a copy of the GNU General Public License         #
#    along with this program; see the file COPYING.  If not, write to          #
#    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,          #
#    Boston, MA 02111-1307, USA.                                               #
#                                                                              #
################################################################################

*/

