/*
  interact: model fault interactions using dislocations in a 
            halfspace
  (C) Thorsten Becker, becker@eps.harvard.edu

  $Id: print_patch_geometry.c,v 1.23 2003/01/07 02:52:29 tbecker Exp $

  output of patch gemeotry and other stuff to *out stream
  in various formats

*/
#include "interact.h"
#include "properties.h"


void print_patch_geometry_and_bc(int flt,struct flt *fault,
				 int opmode,
				 COMP_PRECISION stop_time,
				 my_boolean calculate_base_vecs,
				 FILE *out,
				 my_boolean shrink_patches)
{
  int k,l;
  static my_boolean bc_init=FALSE,g_init=FALSE;
  static int nrf,bc_code;
  COMP_PRECISION corner[4][3],alpha,sin_dip,cos_dip,x[3],lfac,leeway;
#ifdef ALLOW_NON_3DQUAD_GEOM
  COMP_PRECISION tmpdbl;
#endif
  //
  // shrink patches for easier viewing
  //
  if(shrink_patches)
    leeway = 0.9;
  else
    leeway = 1.0;

  if(calculate_base_vecs){
#ifdef ALLOW_NON_3DQUAD_GEOM
    if(fault[flt].type != TRIANGULAR){
      // normal (rectangular, 2-D, or point source)
      alpha=90.0-(COMP_PRECISION)fault[flt].strike;
      my_sincos_deg(&fault[flt].sin_alpha,&fault[flt].cos_alpha,(COMP_PRECISION)alpha);
      my_sincos_deg(&sin_dip,&cos_dip,(COMP_PRECISION)fault[flt].dip);
      calc_base_vecs(fault[flt].t_strike,fault[flt].normal,fault[flt].t_dip,
		     fault[flt].sin_alpha,fault[flt].cos_alpha,sin_dip,cos_dip);
    }else{// triangular element
      get_alpha_dip_tri_gh((fault+flt)->xt,&(fault+flt)->sin_alpha,
			   &(fault+flt)->cos_alpha,&tmpdbl,&(fault+flt)->w);
      (fault+flt)->dip=(float)tmpdbl;
      (fault+flt)->area = (fault+flt)->w;
      alpha=RAD2DEGF(asin((fault+flt)->sin_alpha));
      (fault+flt)->strike= 90.0 - alpha;
    }
#else
    //
    // rectangular patch
    //
    alpha=90.0-(COMP_PRECISION)fault[flt].strike;
    my_sincos_deg(&fault[flt].sin_alpha,&fault[flt].cos_alpha,(COMP_PRECISION)alpha);
    my_sincos_deg(&sin_dip,&cos_dip,(COMP_PRECISION)fault[flt].dip);
    calc_base_vecs(fault[flt].t_strike,fault[flt].normal,fault[flt].t_dip,
		   fault[flt].sin_alpha,fault[flt].cos_alpha,sin_dip,cos_dip);
#endif
  }
  switch(opmode){
  case PATCH_OUT_MODE:{
    /*
      
      output in the PATCH format as read by interact

    */
#ifdef ALLOW_NON_3DQUAD_GEOM
    switch(fault[flt].type){
    case POINT_SOURCE:{
      // L' = -aspect_ratio = - L/W , W' = area = 4 L W
      fprintf(out,"%19.12e %19.12e %19.12e %10.6f %10.6f %19.12e %19.12e %6i\n",
	      fault[flt].x[X], fault[flt].x[Y], 
	      fault[flt].x[Z],fault[flt].strike,
	      fault[flt].dip,-fault[flt].l/fault[flt].w,
	      4.0*fault[flt].w*fault[flt].l,fault[flt].group);
      break;
    }
    case TWO_DIM_SEGMENT_PLANE_STRAIN:
    case TWO_DIM_SEGMENT_PLANE_STRESS:
    case RECTANGULAR_PATCH:{// normal rectangular fault, this is the default
      fprintf(out,"%19.12e %19.12e %19.12e %10.6f %10.6f %19.12e %19.12e %6i\n",
	      fault[flt].x[X], fault[flt].x[Y], 
	      fault[flt].x[Z],fault[flt].strike,
	      fault[flt].dip,
	      fault[flt].l,fault[flt].w,fault[flt].group);
      break;
    }
    case TRIANGULAR:{// xt has to be assigned and allocated before
      fprintf(out,"%15.8e %15.8e %15.8e %10.6f %10.6f %15.8e %15.8e %6i ",
	      999.,999.,999.,999.,999.,-1.,-1.,fault[flt].group);
      fprintf(out,"%15.8e %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e %15.8e\n",
	      fault[flt].xt[X],fault[flt].xt[Y],fault[flt].xt[Z],
	      fault[flt].xt[3+X],fault[flt].xt[3+Y],fault[flt].xt[3+Z],
      	      fault[flt].xt[6+X],fault[flt].xt[6+Y],fault[flt].xt[6+Z]);
      break;
    }
    default:
      fprintf(stderr,"print_fault_patch: patch type %i undefined\n",
	      fault[flt].type);
      exit(-1);
      break;
    }
#else
    // only rectangular
    fprintf(out,"%15.8e %15.8e %15.8e %10.6f %10.6f %15.8e %15.8e %6i\n",
	    fault[flt].x[X], fault[flt].x[Y], fault[flt].x[Z],fault[flt].strike,
	    fault[flt].dip,fault[flt].l,fault[flt].w,fault[flt].group);
#endif
    break;
  }
  case GEOMVIEW_MODE:{
    /*

      output in GEOMVIEW format for 3-D viewing

     */
    if(!g_init){
      fprintf(out,"QUAD\n");
      g_init=TRUE;
    }
#ifdef ALLOW_NON_3DQUAD_GEOM
    switch(fault[flt].type){
    case POINT_SOURCE:
    case TWO_DIM_SEGMENT_PLANE_STRAIN:
    case TWO_DIM_SEGMENT_PLANE_STRESS:
    case RECTANGULAR_PATCH:{
      calculate_bloated_corners(corner,(fault+flt),leeway);
      for(k=0;k<4;k++){
	for(l=0;l<3;l++)
	  if(fabs(corner[k][l]/CHAR_FAULT_DIM)>EPS_COMP_PREC)
	    fprintf(out,"%g ",corner[k][l]/CHAR_FAULT_DIM);
	  else
	    fprintf(out,"0.0 ");
	fprintf(out,"\n");
      }
      break;
    }
    case TRIANGULAR:{
      for(k=0;k<3;k++)
	for(l=0;l<3;l++)
	  fprintf(out,"%g ",fault[flt].xt[k*3+l]);
      for(k=l=0;l<3;l++)
	fprintf(out,"%g ",fault[flt].xt[k*3+l]);
      fprintf(out,"\n");
      break;
    }}
#else
    // only rectangular faults
    calculate_bloated_corners(corner,(fault+flt),leeway);
    for(k=0;k<4;k++){
      for(l=0;l<3;l++)
	if(fabs(corner[k][l]/CHAR_FAULT_DIM)>
	   EPS_COMP_PREC)
	  fprintf(out,"%g ",corner[k][l]/CHAR_FAULT_DIM);
	else
	  fprintf(out,"0.0 ");
      fprintf(out,"\n");
    }
#endif
    break;
  }
  /*

    output in XYZ format for plotting with pxxyz from
    GMT

  */
  case PSXYZ_MODE:{
#ifdef ALLOW_NON_3DQUAD_GEOM
    switch(fault[flt].type){
    case TWO_DIM_SEGMENT_PLANE_STRAIN:
    case TWO_DIM_SEGMENT_PLANE_STRESS:{
      calculate_bloated_corners(corner,(fault+flt),leeway);
      // draw segment with endbars
      lfac = fault[flt].l * 0.2;
      for(l=0;l<3;l++)
	x[l] = corner[0][l] + fault[flt].normal[l] * lfac;
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      for(l=0;l<3;l++)
	x[l] = corner[0][l] - fault[flt].normal[l] * lfac;
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      for(l=0;l<3;l++)
	x[l] = corner[0][l];
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      for(l=0;l<3;l++)
	x[l] = corner[1][l];
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      for(l=0;l<3;l++)
	x[l] = corner[1][l] + fault[flt].normal[l] * lfac;
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      for(l=0;l<3;l++)
	x[l] = corner[1][l] - fault[flt].normal[l] * lfac;
      fprintf(out,"%g %g %g\n",x[X]/CHAR_FAULT_DIM,x[Y]/CHAR_FAULT_DIM,x[Z]/CHAR_FAULT_DIM);
      break;
    }
    case POINT_SOURCE:
    case RECTANGULAR_PATCH:{
      calculate_bloated_corners(corner,(fault+flt),leeway);
      for(k=0;k<4;k++){
	for(l=0;l<3;l++){
	  if(fabs(corner[k][l]/CHAR_FAULT_DIM)>
	     EPS_COMP_PREC)
	    fprintf(out,"%g ",corner[k][l]/CHAR_FAULT_DIM);
	  else
	    fprintf(out,"0.0 ");
	}
	fprintf(out,"\n");
      }
      break;
    }
    case TRIANGULAR:{
      for(k=0;k<3;k++){
	for(l=0;l<3;l++)
	  fprintf(out,"%g ",fault[flt].xt[k*3+l]);
	fprintf(out,"\n");
      }
      for(k=l=0;l<3;l++)
	fprintf(out,"%g ",fault[flt].xt[k*3+l]);
      fprintf(out,"\n");
      break;
    }
    }
#else
    calculate_bloated_corners(corner,(fault+flt),leeway);
    for(k=0;k<4;k++){
      for(l=0;l<3;l++){
	if(fabs(corner[k][l]/CHAR_FAULT_DIM)>
	   EPS_COMP_PREC)
	  fprintf(out,"%g ",corner[k][l]/CHAR_FAULT_DIM);
	else
	  fprintf(out,"0.0 ");
      }
      fprintf(out,"\n");
    }
#endif
    fprintf(out,">\n");
    break;
  }
  case CORNEROUT_MODE:{
#ifdef ALLOW_POINT_SOUCE
    if((fault[flt].type != POINT_SOURCE)&&
       (fault[flt].type != RECTANGULAR_PATCH)&&
       (!patch_is_2d(fault[flt].type))){
      fprintf(stderr,"print_patch_geometry: corner out defined for rectangular, point source and 2-Dpatches only\n");
      exit(-1);
    }
#endif
    calculate_bloated_corners(corner,(fault+flt),leeway);
    for(k=0;k<4;k++){
      for(l=0;l<3;l++){
	if(fabs(corner[k][l]/CHAR_FAULT_DIM)>
	   EPS_COMP_PREC)
	  fprintf(out,"%g ",corner[k][l]/CHAR_FAULT_DIM);
	else
	  fprintf(out,"0 ");
      }
    }
    fprintf(out,"\n");
    break;
  }
  case BC_OUT_MODE:{
    if(!bc_init){
      fprintf(out,"%i\n",SIMULATE_LOADING_AND_PLOT);
      fprintf(out,"%g %g %g\n",0.01,0.01,stop_time);
      fprintf(out,"%g %g\n",0.01,0.1); 
      bc_init=TRUE;
      nrf=0;
    }
    if(fault[flt].strike > 45 && 
       fault[flt].strike <= 135)
      bc_code=STRIKE_SLIP_RIGHTLATERAL;
    else
      bc_code=STRIKE_SLIP_LEFTLATERAL;
    fprintf(out,"-1 %i %i %i\n",nrf,nrf,bc_code);
    nrf++;
    break;
  }
  case VEC_OUT_MODE:{
    fprintf(out,"%12.5e %12.5e %12.5e\t %12.5e %12.5e %12.5e\t %12.5e %12.5e %12.5e\n",
	   fault[flt].t_strike[X],fault[flt].t_strike[Y],
	   fault[flt].t_strike[Z],
	   fault[flt].t_dip[X],fault[flt].t_dip[Y],
	   fault[flt].t_dip[Z],
	   fault[flt].normal[X],fault[flt].normal[Y],
	   fault[flt].normal[Z]);
    break;
  }
  case XYZ_AND_VEC_MODE:{
#ifdef ALLOW_POINT_SOUCE
    if((fault[flt].type != POINT_SOURCE)&&
       (fault[flt].type != RECTANGULAR_PATCH)&&
       (!patch_is_2d(fault[flt].type))){
      fprintf(stderr,"print_patch_geometry: corner out defined for rectangular, point source and 2-Dpatches only\n");
      exit(-1);
    }
#endif
    calculate_bloated_corners(corner,(fault+flt),leeway);
    for(k=0;k<4;k++){
      for(l=0;l<3;l++){
	if(fabs(corner[k][l]/CHAR_FAULT_DIM)>
	   EPS_COMP_PREC)
	  fprintf(out,"%12.5e ",corner[k][l]/CHAR_FAULT_DIM);
	else
	  fprintf(out,"%12.5e ",0.0);
      }
    }
    fprintf(out,"%12.5e %12.5e %12.5e %12.5e %12.5e %12.5e %12.5e %12.5e %12.5e\n",
	    fault[flt].t_strike[X],fault[flt].t_strike[Y],
	    fault[flt].t_strike[Z],
	    fault[flt].t_dip[X],fault[flt].t_dip[Y],
	    fault[flt].t_dip[Z],
	    fault[flt].normal[X],fault[flt].normal[Y],
	    fault[flt].normal[Z]);
    break;
  }
  default:{
    fprintf(stderr,"print_patch_geometry: mode %i undefined\n",
	    opmode);
    exit(-1);
  }}
}




/*

################################################################################
#                                                                              #
#  interact: model fault interactions using dislocations in an elastic         #
#            medium                                                            #
#	                                                                       #
#    Copyright (C) Thorsten W. Becker 2000 - 2002                              #
#                                                                              #
#    Unteract uses dc3d.f as provided by Y. Okada as in Okada (BSSA, 1992) and #
#    linear algebra routines from SLATEC, LAPACK, and EISPACK.                 #
#    Might contain material copyrighted by others (e.g. Numerical Recipes etc) #
#                                                                              #
#  If you use interact in real work, a citation to use would be:               #
#                                                                              #
#  @Article{becker02c,                                                         #
#    author = 	 {Becker, T. W. and Schott, B.},                               #
#    title = 	 {On boundary-element models of elastic fault interaction      #
#                  (abstract)},                                                # 
#    journal = 	 {EOS Trans. AGU},                                             #
#    year = 	 2002,                                                         #
#    volume =	 83,                                                           #
#    number =	 47,                                                           #
#    pages =	 {NG62A-0925}                                                  #
#  }                                                                           #
#                                                                              #
#                                                                              #
#    This program is free software; you can redistribute it and/or modify      #
#    it under the terms of the GNU General Public License as published by      #
#    the Free Software Foundation; either version 2 of the License, or         #
#    (at your option) any later version.                                       #
#                                                                              #
#    This program is distributed in the hope that it will be useful,           #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of            #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
#    GNU General Public License for more details.                              #
#                                                                              #
#    In addition, we strongly suggest that iGMT users comply with the goals    #
#    as expressed in the Student Pugwash Pledge (www.spusa.org/pugwash/).      #
#                                                                              #
#    You should have received a copy of the GNU General Public License         #
#    along with this program; see the file COPYING.  If not, write to          #
#    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,          #
#    Boston, MA 02111-1307, USA.                                               #
#                                                                              #
################################################################################

*/

