/*
  interact: model fault interactions using dislocations in a 
            halfspace
  (C) Thorsten Becker, becker@eps.harvard.edu

  $Id: quake.c,v 2.25 2003/01/07 02:52:29 tbecker Exp $
*/
#include "interact.h"
#include "properties.h"
/* 
   - add displacement to fault,
   - add coseismic stresses to other faults 
   - write event to file
   - if set, plot quake on X window

   if add_stress is not set, will not add slip and stress change
   if mark_quake is not set, will not add to plotting and moment lists

   sma: activation codes
   slip: slip vector, both are in strike, dip, normal frame
   r_flt: slipping fault number
   

*/


void quake(my_boolean *sma,COMP_PRECISION *slip,int r_flt,struct flt *fault,
	   struct med *medium,my_boolean add_stress, my_boolean mark_quake)
{
#ifdef USE_PGPLOT
  int offset,j;
#endif
  int i;
  float fslip[3];
  COMP_PRECISION cslip[3];
  COMP_PRECISION mom;
  //
  if(!medium->moment_release_init){
    /* first time around, initialize fields */
    medium->moment_release_init=TRUE;
    if(medium->events_init)
      medium->old_moment_time=medium->time;
  }
  if(add_stress)
    add_quake_stress(r_flt,sma,slip,fault,medium);
  /*

    register the slip event in event files and the like (a 'real earthquake')

  */
  if(mark_quake){
    // calculate moment of slipping patch
    mom  = norm_3d(slip) * fault[r_flt].area * SHEAR_MODULUS;
    // write single event to  event files
    if(medium->events_init){
      // single patch 
      for(i=0;i<3;i++)
	fslip[i]=(float)slip[i];
      write_patch_event_file((float)medium->time,medium->iter, r_flt,(float)mom,
			      fslip,medium->events_out);
     }
    // flush group output stack from previouse if we are at a different time
    if(medium->time != medium->old_moment_time)
      flush_moment_stack(medium);
    //
    // add new event to total moment
    medium->total_moment += mom;
    //
    // add moment to temporary group moment
    medium->fault_group[fault[r_flt].group].moment += mom;
    //
    // add weighted slip to temporary group slip
    medium->fault_group[fault[r_flt].group].sarea += 
      fault[r_flt].area;
    /* convert to cartesian */
    compute_cartesian_slip(cslip,slip,(fault+r_flt));
    for(i=0;i<3;i++){
      /* centroid location */
      medium->fault_group[fault[r_flt].group].cent[i]   += fault[r_flt].x[i]  * fault[r_flt].area;
      /* strike,dip,normal */
      medium->fault_group[fault[r_flt].group].slip[i]   += slip[i]  * fault[r_flt].area;
      /* x,y,z */
      medium->fault_group[fault[r_flt].group].slip[3+i] += cslip[i] * fault[r_flt].area;
    }
#ifdef USE_PGPLOT
    if(medium->moment_array_init){
      i=(int)((fault[r_flt].pos[1]+0.5)*
	      (COMP_PRECISION)(medium->grp0_n-1)+0.5);
      j=(int)((fault[r_flt].pos[0]+0.5)*
	      (COMP_PRECISION)(medium->grp0_m-1)+0.5);
      offset  = i*medium->grp0_m+j;
      medium->momrel[offset] += mom;
    }
    plot_quake(r_flt,medium,fault);
#endif
  }
}
/*

  deal with the stress and slip increments due to slip on a patch,
  i.e. add it's effect on stress on other faults

  
*/
void add_quake_stress(int r_flt, my_boolean *sma,
		      COMP_PRECISION *slip, 
		      struct flt *fault, struct med *medium)
{
  int i;
  /* 
     add the stress and slip to the fault patch 
     and other that are affected
     
     add slip to rupturing fault 
  */
  for(i=0;i<3;i++)
    if(sma[i])
      fault[r_flt].u[i] += slip[i];
  /* 
 
     add stress contributions of rupture at fault r_flt with slip[] to all other 
     affected faults 

  */
  switch(select_i_coeff_calc_mode(medium)){
  case I_MAT_IN_MEMORY:{
    add_quake_stress_1(sma,slip,r_flt,fault,medium);
    break;
  }
  case I_MAT_ON_FILE:{
    add_quake_stress_2(sma,slip,r_flt,fault,medium);
    break;
  }
  case CALC_I_COEFF_NOW:{
    add_quake_stress_3(sma,slip,r_flt,fault,medium);
    break;
  }
  case SPARSE_I_MAT:{
    add_quake_stress_4(sma,slip,r_flt,fault,medium);
    break;
  }
  default:{
    fprintf(stderr,"add_quake_stress: can not deal with I matrix code %i\n",
	    select_i_coeff_calc_mode(medium));
    exit(-1);
  }}

}

/*

################################################################################
#                                                                              #
#  interact: model fault interactions using dislocations in an elastic         #
#            medium                                                            #
#	                                                                       #
#    Copyright (C) Thorsten W. Becker 2000 - 2002                              #
#                                                                              #
#    Unteract uses dc3d.f as provided by Y. Okada as in Okada (BSSA, 1992) and #
#    linear algebra routines from SLATEC, LAPACK, and EISPACK.                 #
#    Might contain material copyrighted by others (e.g. Numerical Recipes etc) #
#                                                                              #
#  If you use interact in real work, a citation to use would be:               #
#                                                                              #
#  @Article{becker02c,                                                         #
#    author = 	 {Becker, T. W. and Schott, B.},                               #
#    title = 	 {On boundary-element models of elastic fault interaction      #
#                  (abstract)},                                                # 
#    journal = 	 {EOS Trans. AGU},                                             #
#    year = 	 2002,                                                         #
#    volume =	 83,                                                           #
#    number =	 47,                                                           #
#    pages =	 {NG62A-0925}                                                  #
#  }                                                                           #
#                                                                              #
#                                                                              #
#    This program is free software; you can redistribute it and/or modify      #
#    it under the terms of the GNU General Public License as published by      #
#    the Free Software Foundation; either version 2 of the License, or         #
#    (at your option) any later version.                                       #
#                                                                              #
#    This program is distributed in the hope that it will be useful,           #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of            #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
#    GNU General Public License for more details.                              #
#                                                                              #
#    In addition, we strongly suggest that iGMT users comply with the goals    #
#    as expressed in the Student Pugwash Pledge (www.spusa.org/pugwash/).      #
#                                                                              #
#    You should have received a copy of the GNU General Public License         #
#    along with this program; see the file COPYING.  If not, write to          #
#    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,          #
#    Boston, MA 02111-1307, USA.                                               #
#                                                                              #
################################################################################

*/

